

#include "TriMRFEnergy.h"



TriMRFEnergy::TriMRFEnergy(int v_num, int f_num, int l_num): _V(v_num), _F(f_num), _L(l_num)
{
	arrayVert = new float[_V * _L];
	arrayFace = new float[_F * _L * _L * _L];
	fIdxArray = new int[_F * 3];
	
	for(int v = 0; v < _V; v ++)
		v_array.push_back(vertex(_L, arrayVert + v *_L));
	
	for(int f = 0; f < _F; f ++) {
		f_array.push_back(face(_L, arrayFace + f *_L * _L * _L));
		fIdxArray[3*f + 0] = 0;
		fIdxArray[3*f + 1] = 1;
		fIdxArray[3*f + 2] = 2;
	}
}

TriMRFEnergy::~TriMRFEnergy()
{
	delete[] arrayVert;
	delete[] arrayFace;
	delete[] fIdxArray;
}



int TriMRFEnergy::vertex::min_label()
{
	int min_l = 0;
	float min_val = 1e10;
	for (int i = 0; i < _L; i++) 
		if (min_val > theta_single[i]) {
			min_val = theta_single[i];
			min_l = i;
		}
	return min_l;
}

void TriMRFEnergy::AddFaceInd2Nd(int fID, int vID1, int vID2, int vID3)
{
	f_array[fID].vID[0] = vID1;
	f_array[fID].vID[1] = vID2;
	f_array[fID].vID[2] = vID3;
	fIdxArray[3*fID + 0] = vID1;
	fIdxArray[3*fID + 1] = vID2;
	fIdxArray[3*fID + 2] = vID3;
}


void TriMRFEnergy::AddFaceData(int fID, float *data)
{
	for (int i = 0; i < _L*_L*_L; i ++ ) 
		f_array[fID].theta_tri[i] = data[i];
}

void TriMRFEnergy::AddNodeData(int nID, float *data)
{
	for (int i = 0; i < _L; i ++) 
		v_array[nID].theta_single[i] = data[i];
}

// a priority queue implementation is more efficient
// is there an O(n) impl.?
void TriMRFEnergy::randPermutation(int n, std::vector<int> &p_array)
{
	p_array.clear();
	int k = 0;
	int *indicator = new int[n];
	for (int i = 0; i < n; i ++) indicator[i] = 0;
	while ( k < n) {
		int idx = (int)(((float)rand()/RAND_MAX)*(n-k));
		int cnt = 0;
		for (int i = 0; i < n; i++) {
			if ( indicator[i] == 0 ) {
				if (cnt == idx) {
					indicator[i] = 1; 
					p_array.push_back(i);
					break;
				} else 
					cnt ++;
			}
		}
		k++;
	}
}

void TriMRFEnergy::LoadCudaData()
{	
#ifdef CUDA_OPTIMIZE
	cudaLoadEnergy(_V, _F, _L, arrayVert, arrayFace, fIdxArray);
/*	for (int i = 0; i < _F*_L*_L*_L; i++) {
		printf("__f: %d -- %f\n", i, arrayFace[i]);
	}
	getchar();*/
#endif
}

void TriMRFEnergy::ClearCudaData()
{
#ifdef CUDA_OPTIMIZE
	cudaClearTriMRF();
#endif
}

int TriMRFEnergy::optimize(Options &options)
{
	switch ( options.optimizer ) {
		case CPU_RANDOM_ORDER:
			return optimize_CPU_RandOrder(options);		
		case CPU_FIXED_ORDER:
			return optimize_CPU_FixedOrder(options);		
		case GPU:		
			return optimize_GPU(options);
		case CPU_RANDOM_PERFACE:
			return optimize_CPU_RandPerFace(options);
	}
	return 0;
}