// includes, system
#include <stdio.h>
#include <assert.h>
#include <vector>
#include "TriMRFEnergy.h"

/*	
*/

// Simple utility function to check for CUDA runtime errors
void checkCUDAErrorTriMRF(const char *msg);

extern "C" void cudaLoadEnergy (int V, int F, int L, float *theta_single, float *theta_tri, int *fIdxArray) ;
extern "C" void cudaGetEnergy (int V, int F, int L, float *theta_single, float *theta_tri);
extern "C" void cudaOptimizeTriMRF (int V, int F, int L, float eps, int maxItr);
extern "C" void cudaOptimizeTriMRFNew (int V, int F, int L, float eps, int maxItr, std::vector<std::vector<int> > &ff_list);
extern "C" void cudaClearTriMRF ();

float* d_array_single = NULL, *d_array_tri = NULL;
float* d_array_indicator = NULL;
float* h_array_indicator = NULL;
float* h_array_indicator_org = NULL;
int * d_array_fIdx = NULL;

void cudaLoadEnergy (int V, int F, int L, 
					 float *theta_single, float *theta_tri, 
					 int *fIdxArray)
{	
    if (d_array_single != NULL)
		cudaFree(d_array_single);
	if (d_array_tri != NULL)
		cudaFree(d_array_tri);
	if (d_array_indicator != NULL)
		cudaFree(d_array_indicator);
	if (h_array_indicator != NULL)
		delete[] h_array_indicator;
	if (h_array_indicator_org != NULL)
		delete[] h_array_indicator_org;
	if (d_array_fIdx != NULL)
		cudaFree(d_array_fIdx);
	
    cudaMalloc( (void**) &d_array_single, L*V*sizeof(float));	
	cudaMemcpy(d_array_single, theta_single, L*V*sizeof(float), cudaMemcpyHostToDevice);
	cudaMalloc( (void**) &d_array_tri, L*L*L*F*sizeof(float));	
	cudaMemcpy(d_array_tri, theta_tri, L*L*L*F*sizeof(float), cudaMemcpyHostToDevice);
	cudaMalloc( (void**) &d_array_indicator, F*sizeof(float));	
	cudaMalloc( (void**) &d_array_fIdx, 3*F*sizeof(int));	
	cudaMemcpy(d_array_fIdx, fIdxArray, 3*F*sizeof(int), cudaMemcpyHostToDevice);
	

	h_array_indicator = new float[F];
	h_array_indicator_org = new float[F];
	for (int i = 0; i < F; i ++) h_array_indicator_org[i] = 0;	

}

void cudaGetEnergy (int V, int F, int L, float *theta_single, float *theta_tri)
{	
    if (d_array_single == NULL || d_array_tri == NULL)
		return;
    cudaMemcpy(theta_single, d_array_single, L*V*sizeof(float), cudaMemcpyDeviceToHost);
	cudaMemcpy(theta_tri, d_array_tri, L*L*L*F*sizeof(float), cudaMemcpyDeviceToHost);	
}

void checkCUDAErrorTriMRF(const char *msg)
{
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err) 
    {
        fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err) );
        exit(-1);
    }                         
}

__device__ float min_jk(float *theta_tri, int baseIdx, int v, int l, int _V, int _L)
{
	float min_val = 1e10;
	switch (v) {
	case 0:
		for (int i = l; i < _L*_L*_L; i += _L) {
			if (min_val > theta_tri[baseIdx + i])
				min_val = theta_tri[baseIdx + i];
		}
		break;
	case 1:
		for (int i = l*_L; i < _L*_L*_L; i += _L*_L)
			for (int j = 0; j < _L; j++) {
				if (min_val > theta_tri[baseIdx + i])
					min_val = theta_tri[baseIdx + i];
			}
		break;
	case 2:
		for (int i = l*_L*_L; i < (l+1)*_L*_L; i ++)	 {
			if (min_val > theta_tri[baseIdx + i])
				min_val = theta_tri[baseIdx + i];
		}
		break;
	}
	return min_val;
}

__device__ void add_msg(float *theta_tri, int baseIdx, int v, int l, float w_diff, int _V, int _L)
{
	switch (v) {
	case 0:
		for (int i = l; i < _L*_L*_L; i += _L)
				theta_tri[baseIdx + i] += w_diff;
		break;
	case 1:
		for (int i = l*_L; i < _L*_L*_L; i += _L*_L)
			for (int j = 0; j < _L; j++) 			
				theta_tri[baseIdx + i+j] += w_diff;
		break;
	case 2:
		for (int i = l*_L*_L; i < (l+1)*_L*_L; i ++)	
				theta_tri[baseIdx + i] += w_diff;
		break;
	}
}

__global__ void IterateFace(int fID, int V, int F, int L, float eps, 
							float *theta_single, float *theta_tri, 
							float *array_indicator, int *fIdxArray)
{
	int lIdx = threadIdx.x;//blockIdx.x * blockDim.x + threadIdx.x;	
	//array_indicator[fID] = lIdx;//threadIdx.x;///max((float)threadIdx.z,array_indicator[fID]);
	//return;
	if (lIdx >= L)
		return;
	
	
	int vid[3];
	vid[0] = fIdxArray[3*fID + 0]*L;
	vid[1] = fIdxArray[3*fID + 1]*L;
	vid[2] = fIdxArray[3*fID + 2]*L;
	int fIdx = fID * L * L * L;
	int l = lIdx;
	int itr = 0;
	while(itr++ < 100) {
		for (int v = 0; v < 3; v++) {
			//for (int l = 0; l < L; l ++) {
				float min_v = min_jk(theta_tri, fIdx, v, l, V, L);
				float dif = 0.5 * (theta_single[vid[v]+l] - min_v);
				theta_single[vid[v]+l] -= dif;
				add_msg(theta_tri, fIdx, v, l, dif, V, L);
				if (dif > eps) {
					array_indicator[fID] = 1;
				}
				//array_indicator[fID] = max((float)l,array_indicator[fID]);
			//}
		}
	}
	//array_indicator[fID] = fID;
	//__syncthreads();
}

__global__ void IterateFaceNew(int fMax, int v, int *fListArr,// list for face set
							   int V, int F, int L, float eps, 
							   float *theta_single, float *theta_tri, 
							   float *array_indicator, int *fIdxArray)// List for vertex index
{
	int lIdx = threadIdx.x;	
	int fIdx = blockIdx.x;
	
	//array_indicator[0] = fIdx;//atomicMax((float)lIdx,array_indicator[fID]);
	//return;
	if (lIdx >= L || fIdx >= fMax)
		return;
	
	int fID = fListArr[fIdx];	
	int vid[3];
	vid[0] = fIdxArray[3*fID + 0]*L;
	vid[1] = fIdxArray[3*fID + 1]*L;
	vid[2] = fIdxArray[3*fID + 2]*L;

	int fBaseAddr = fID * L * L * L;
	int l = lIdx;	


	float min_v = min_jk(theta_tri, fBaseAddr, v, l, V, L);
	float dif = 0.5 * (theta_single[vid[v]+l] - min_v);
	theta_single[vid[v]+l] -= dif;
	add_msg(theta_tri, fBaseAddr, v, l, dif, V, L);
	if (dif > eps) {
		array_indicator[fID] = 1;
	}
	
}



void cudaOptimizeTriMRFNew(int V, int F, int L, float eps, int maxItr, std::vector< std::vector<int> > &ff_list)
{
	if (d_array_single == NULL || d_array_tri == NULL) {
		printf("Please load energy first!\n");
		return;
	}

	// block size for parallel computation
	const int maxLabelSize = 64; // maximal label size: 64
    dim3 dimBlock(maxLabelSize, 1, 1 ); //
	dim3 dimGrid(1, 1, 1);	// grid can be at most 2D

	// Timers
	cudaEvent_t start, stop; 
    float time;
    cudaEventCreate(&start);
	cudaEventCreate(&stop);


	int max_size = ff_list[0].size()*2;
	int *fListArr = new int[max_size];
	int *d_fListArr = NULL;
	cudaMalloc( (void**) &d_fListArr, max_size*sizeof(int));	

	// Being iterations
	int itr = 0;
	bool is_done = false;
	while (itr ++ < maxItr && !is_done) {
		//if (itr % 5 == 0)
		printf("iteration: %d ...\n", itr);
		// reset zero
		cudaMemcpy(d_array_indicator, h_array_indicator_org, F*sizeof(float), cudaMemcpyHostToDevice);
		cudaThreadSynchronize();

		// iterate for all the faces
		cudaEventRecord( start, 0 );
		/////////////////////////////////////////////////////////////////////
		for(int idx = 0; idx < ff_list.size(); idx ++) {
			dimGrid.x = ff_list[idx].size();
			//printf("face set %d, no. %d \n", idx, ff_list[idx].size());
			for (int k = 0; k < ff_list[idx].size(); k++) fListArr[k] = ff_list[idx][k];
			cudaMemcpy(d_fListArr, fListArr, ff_list[idx].size()*sizeof(int), cudaMemcpyHostToDevice);

			for (int v = 0; v < 3; v ++) {
				IterateFaceNew<<<dimGrid, dimBlock>>>(ff_list[idx].size(), v, d_fListArr,
					V, F, L, eps, d_array_single, d_array_tri, d_array_indicator, 
					d_array_fIdx);
				cudaThreadSynchronize();
			}
		}
		/////////////////////////////////////////////////////////////////////
		cudaEventRecord( stop, 0 ); 
		cudaEventSynchronize( stop );
		cudaEventElapsedTime( &time, start, stop );
		printf("Elapsed Time: %f ms\n", time);		

		//if (itr % 5 == 0) {
		is_done = true;
		cudaMemcpy(h_array_indicator, d_array_indicator, F*sizeof(float), cudaMemcpyDeviceToHost);
		cudaThreadSynchronize();
		for (int f = 0; f < F; f++) {
			//printf("f: %d -- %f\n", f, h_array_indicator[f]);
			if(h_array_indicator[f] > 1e-3) {
				is_done = false;
				break;
			}
		}
/*		float *p_array = new float[F*L*L*L];
		cudaMemcpy(p_array, d_array_tri, F*L*L*L*sizeof(float), cudaMemcpyDeviceToHost);
		for (int i = 0; i < F*L*L*L; i++) {
			printf("f: %d -- %f\n", i, p_array[i]);
		}
		delete[] p_array;
		getchar();*/
	}


	// Destroy timers
	// destroy event
	cudaEventDestroy(start);
	cudaEventDestroy(stop);

}

void cudaClearTriMRF()
{
	if (d_array_single != NULL)
		cudaFree(d_array_single);
	if (d_array_tri != NULL)
		cudaFree(d_array_tri);
	if (d_array_indicator != NULL)
		cudaFree(d_array_indicator);
	if (h_array_indicator != NULL)
		delete[] h_array_indicator;
	if (h_array_indicator_org != NULL)
		delete[] h_array_indicator_org;
	if (d_array_fIdx != NULL)
		cudaFree(d_array_fIdx);
	d_array_tri = d_array_single = d_array_indicator = 
		h_array_indicator = h_array_indicator_org = NULL;
	d_array_fIdx = NULL;
}



void cudaOptimizeTriMRF(int V, int F, int L, float eps, int maxItr)
{
	if (d_array_single == NULL || d_array_tri == NULL) {
		printf("Please load energy first!\n");
		return;
	}

	// block size for parallel computation
	const int maxLabelSize = 64; // maximal label size: 64
    dim3 dimBlock(maxLabelSize, 1, 1 ); //
	dim3 dimGrid(1, 1, 1);	// grid can be at most 2D
	
	// Timers
	cudaEvent_t start, stop; 
    float time;
    cudaEventCreate(&start);
	cudaEventCreate(&stop);


	// Being iterations
	int itr = 0;
	bool is_done = false;
	while (itr ++ < maxItr && !is_done) {
		is_done = false;
		//if (itr % 5 == 0)
		printf("iteration: %d ...\n", itr);
		// reset zero
		cudaMemcpy(d_array_indicator, h_array_indicator_org, F*sizeof(float), cudaMemcpyHostToDevice);

		cudaEventRecord( start, 0 );
		/////////////////////////////////////////////////////////////////////
		for (int f = 0; f < F; f++) 
			IterateFace<<<dimGrid, dimBlock>>>(f, 
				V, F, L, eps, 
				d_array_single, d_array_tri, d_array_indicator, d_array_fIdx);
		/////////////////////////////////////////////////////////////////////
		cudaEventRecord( stop, 0 ); 
		cudaEventSynchronize( stop );
		cudaEventElapsedTime( &time, start, stop );
		printf("Elapsed Time: %f ms\n", time);

		//if (itr % 5 == 0) {
		is_done = true;
		cudaMemcpy(h_array_indicator, d_array_indicator, F*sizeof(float), cudaMemcpyDeviceToHost);
		for (int f = 0; f < F; f++) {
			printf("f: %d -- %f\n", f, h_array_indicator[f]);
			if(h_array_indicator[f] > 1e-3) {
				is_done = false;
				break;
			}
		}
			//getchar();
		//}
/*		float *p_array = new float[F*L*L*L];
		cudaMemcpy(p_array, d_array_tri, F*L*L*L*sizeof(float), cudaMemcpyDeviceToHost);
		for (int i = 0; i < F*L*L*L; i++) {
			printf("f: %d -- %f\n", i, p_array[i]);
		}
		delete[] p_array;
		getchar();*/
	}
}
